# Import a model
import laser_sql_model.sir_sql as model
from copy import deepcopy

import settings
import laser_sql_model.report as report

report.write_report = True # sometimes we want to turn this off to check for non-reporting bottlenecks
fractions = {}
report_births = {}
report_deaths = {}

def collect_and_report(csvwriter, timestep):
    currently_infectious, currently_sus, cur_reco = model.collect_report( ctx )
    counts = {
            "S": deepcopy( currently_sus ),
            "I": deepcopy( currently_infectious ),
            "R": deepcopy( cur_reco ) 
        }
    #print( f"Counts =\nS:{counts['S']}\nI:{counts['I']}\nR:{counts['R']}" )
    def normalize( sus, inf, rec ):
        totals = {}
        for idx in currently_sus.keys():
            totals[ idx ] = sus[ idx ] + inf[ idx ] + rec[ idx ]
            if totals[ idx ] > 0:
                sus[ idx ] /= totals[ idx ] 
                inf[ idx ] /= totals[ idx ] 
                rec[ idx ]/= totals[ idx ] 
            else:
                sus[ idx ] = 0
                inf[ idx ] = 0
                rec[ idx ] = 0
        return totals
    totals = normalize( currently_sus, currently_infectious, cur_reco )
    fractions = {
            "S": currently_sus,
            "I": currently_infectious,
            "R": cur_reco 
        }
    #print( fractions["S"][10] )
    #print( counts["S"][10] )
    try:
        report.write_timestep_report( csvwriter, timestep, counts["I"], counts["S"], counts["R"], new_births={}, new_deaths={} )
    except Exception as ex:
        raise ValueError( f"Exception {ex} at timestep {timestep} and counts {counts['I']}, {counts['S']}, {counts['R']}" )
    return counts, fractions, totals

def run_simulation(ctx, csvwriter, num_timesteps):
    counts, fractions, totals = collect_and_report(csvwriter,0)
    for timestep in range(1, num_timesteps + 1):

        # We should always be in a low prev setting so this should only really ever operate
        # on ~1% of the active population
        ctx = model.progress_infections( ctx )

        # The perma-immune should not consume cycles but there could be lots of waning immune
        # ctx = model.progress_immunities( ctx )

        # The core transmission part begins
        new_infections = model.calculate_new_infections( ctx, fractions["I"], fractions["S"], totals )
        #print( f"new_infections=\n{new_infections}" )

        # TBD: for loop should probably be implementation-specific
        ctx = model.handle_transmission( ctx, new_infections )

        ctx = model.add_new_infections( ctx )

        ctx = model.distribute_interventions( ctx, timestep )
        # Transmission is done, now migrate some. Only infected?
        if settings.num_nodes>1:
            ctx = model.migrate( ctx, timestep, num_infected=sum(fractions["I"].values()) )

        # if we have had total fade-out, inject imports
        if sum(counts["I"].values()) == 0:
            print( f"Injecting 100 new cases into node {settings.num_nodes-1}." )
            model.inject_cases( ctx, import_cases=100, import_node=settings.num_nodes-1 )

        # We almost certainly won't waste time updating everyone's ages every timestep but this is 
        # here as a placeholder for "what if we have to do simple math on all the rows?"
        global report_births, report_deaths 
        ( report_births, report_deaths ) = model.update_ages( ctx, totals, timestep )

        # Report
        #currently_infectious, currently_sus, cur_reco = model.collect_report( ctx )
        #totals = normalize( currently_sus, currently_infectious, cur_reco )
        counts, fractions, totals = collect_and_report(csvwriter,timestep)
        #report.write_timestep_report( csvwriter, timestep, currently_infectious, currently_sus, cur_reco )

    print(f"Simulation completed. Report saved to '{settings.report_filename}'.")

# Main simulation
if __name__ == "__main__":
    # Initialize the 'database' (or load the dataframe/csv)
    # ctx might be db cursor or dataframe or dict of numpy vectors
    ctx = model.initialize_database( from_file=False )
    #ctx = model.init_db_from_csv( settings )
    ctx = model.eula_init( ctx, settings.eula_age, eula_strategy="separate" )

    csv_writer = report.init()

    # Run the simulation for 1000 timesteps
    from timeit import timeit
    from functools import partial
    runsim = partial( run_simulation, ctx=ctx, csvwriter=csv_writer, num_timesteps=settings.duration )
    runtime = timeit( runsim, number=1 )
    print( f"Execution time = {runtime}." )

